@@ -12,6 +12,7 @@ from operator import methodcaller
 
 from mock import Mock
 from pytest import mark, raises
+
 from schema import (
     And,
     Const,
@@ -1080,10 +1081,89 @@ def test_json_schema_forbidden_key_ignored():
     }
 
 
-def test_json_schema_not_a_dict():
-    s = Schema(int)
-    with raises(ValueError):
-        s.json_schema("my-id")
+@mark.parametrize(
+    "input_schema, ignore_extra_keys, additional_properties",
+    [
+        ({}, False, False),
+        ({str: str}, False, True),
+        ({Optional(str): str}, False, True),
+        ({object: int}, False, True),
+        ({}, True, True),
+    ],
+)
+def test_json_schema_additional_properties(input_schema, ignore_extra_keys, additional_properties):
+    s = Schema(input_schema, ignore_extra_keys=ignore_extra_keys)
+    assert s.json_schema("my-id") == {
+        "$schema": "http://json-schema.org/draft-07/schema#",
+        "$id": "my-id",
+        "required": [],
+        "properties": {},
+        "additionalProperties": additional_properties,
+        "type": "object",
+    }
+
+
+def test_json_schema_additional_properties_multiple():
+    s = Schema({"named_property": bool, object: int})
+    assert s.json_schema("my-id") == {
+        "$schema": "http://json-schema.org/draft-07/schema#",
+        "$id": "my-id",
+        "required": ["named_property"],
+        "properties": {"named_property": {"type": "boolean"}},
+        "additionalProperties": True,
+        "type": "object",
+    }
+
+
+@mark.parametrize(
+    "input_schema, expected_keyword, expected_value",
+    [
+        (int, "type", "integer"),
+        (float, "type", "number"),
+        (list, "type", "array"),
+        (bool, "type", "boolean"),
+        (dict, "type", "object"),
+        ("test", "const", "test"),
+        (Or(1, 2, 3), "enum", [1, 2, 3]),
+        (Or(str, int), "anyOf", [{"type": "string"}, {"type": "integer"}]),
+        (And(str, "value"), "allOf", [{"type": "string"}, {"const": "value"}]),
+    ],
+)
+def test_json_schema_root_not_dict(input_schema, expected_keyword, expected_value):
+    """Test generating simple JSON Schemas where the root element is not a dict"""
+    json_schema = Schema(input_schema).json_schema("my-id")
+
+    assert json_schema == {
+        expected_keyword: expected_value,
+        "$id": "my-id",
+        "$schema": "http://json-schema.org/draft-07/schema#",
+    }
+
+
+@mark.parametrize(
+    "input_schema, expected_keyword, expected_value",
+    [([1, 2, 3], "enum", [1, 2, 3]), ([1], "const", 1), ([str], "type", "string")],
+)
+def test_json_schema_array(input_schema, expected_keyword, expected_value):
+    json_schema = Schema(input_schema).json_schema("my-id")
+
+    assert json_schema == {
+        "type": "array",
+        "items": {expected_keyword: expected_value},
+        "$id": "my-id",
+        "$schema": "http://json-schema.org/draft-07/schema#",
+    }
+
+
+def test_json_schema_regex_root():
+    json_schema = Schema(Regex("^v\\d+")).json_schema("my-id")
+
+    assert json_schema == {
+        "type": "string",
+        "pattern": "^v\\d+",
+        "$id": "my-id",
+        "$schema": "http://json-schema.org/draft-07/schema#",
+    }
 
 
 def test_json_schema_dict_type():
@@ -1233,6 +1313,23 @@ def test_description_with_default():
     assert s.validate({}) == {"test1": {}}
 
 
+def test_json_schema_ref_in_list():
+    s = Schema(
+        Or(Schema([str], name="Inner test", as_reference=True), Schema([str], name="Inner test2", as_reference=True))
+    )
+    generated_json_schema = s.json_schema("my-id")
+
+    assert generated_json_schema == {
+        "definitions": {
+            "Inner test": {"items": {"type": "string"}, "type": "array"},
+            "Inner test2": {"items": {"type": "string"}, "type": "array"},
+        },
+        "anyOf": [{"$ref": "#/definitions/Inner test"}, {"$ref": "#/definitions/Inner test2"}],
+        "$id": "my-id",
+        "$schema": "http://json-schema.org/draft-07/schema#",
+    }
+
+
 def test_json_schema_refs():
     s = Schema({"test1": str, "test2": str, "test3": str})
     hashed = "#" + str(hash(repr(sorted({"type": "string"}.items()))))
@@ -1326,11 +1423,17 @@ def test_json_schema_definitions():
 
 
 def test_json_schema_definitions_and_literals():
-    sub_schema = Schema({Literal("sub_key1", description="Sub key 1"): int}, name="sub_schema", as_reference=True)
+    sub_schema = Schema(
+        {Literal("sub_key1", description="Sub key 1"): int},
+        name="sub_schema",
+        as_reference=True,
+        description="Sub Schema",
+    )
     main_schema = Schema(
         {
             Literal("main_key1", description="Main Key 1"): str,
             Literal("main_key2", description="Main Key 2"): sub_schema,
+            Literal("main_key3", description="Main Key 3"): sub_schema,
         }
     )
 
@@ -1339,14 +1442,16 @@ def test_json_schema_definitions_and_literals():
         "type": "object",
         "properties": {
             "main_key1": {"description": "Main Key 1", "type": "string"},
-            "main_key2": {"$ref": "#/definitions/sub_schema"},
+            "main_key2": {"$ref": "#/definitions/sub_schema", "description": "Main Key 2"},
+            "main_key3": {"$ref": "#/definitions/sub_schema", "description": "Main Key 3"},
         },
-        "required": ["main_key1", "main_key2"],
+        "required": ["main_key1", "main_key2", "main_key3"],
         "additionalProperties": False,
         "$id": "my-id",
         "$schema": "http://json-schema.org/draft-07/schema#",
         "definitions": {
             "sub_schema": {
+                "description": "Sub Schema",
                 "type": "object",
                 "properties": {"sub_key1": {"description": "Sub key 1", "type": "integer"}},
                 "required": ["sub_key1"],
@@ -1495,3 +1600,18 @@ def test_prepend_schema_name():
         Schema(int, name="custom_schemaname").validate("a")
     except SchemaUnexpectedTypeError as e:
         assert str(e) == "'custom_schemaname' 'a' should be instance of 'int'"
+
+
+def test_dict_literal_error_string():
+    # this is a simplified regression test of the bug in github issue #240
+    assert Schema(Or({"a": 1}, error="error: {}")).is_valid(dict(a=1))
+
+
+def test_callable_error():
+    # this tests for the behavior desired in github pull request #238
+    e = None
+    try:
+        Schema(lambda d: False, error="{}").validate("This is the error message")
+    except SchemaError as ex:
+        e = ex
+    assert e.errors == ["This is the error message"]
